#define TRIANGLENOISE
#define NORMALMAP
#define IBL
#define SHADOW
struct Light {
    vec3 direction;
    vec3 intensity;
    float attenuation;
};

varying vec2 vTexcoord;
varying vec3 vNormal;
varying vec3 vWorldPosition;
varying vec3 vTangent;

uniform sampler2D diffuseMap;
uniform sampler2D normalMap;
uniform sampler2D mraMap;

uniform vec3 cameraPosition;

uniform float fresnelIntense;

uniform float specularIntensity;
uniform float wrapped;
uniform float opacity;

#ifdef MESH_MASK
varying float vMask;
#endif


#define MEDIUMP_FLT_MAX    65504.0
#define saturateMediump(x) min(x, MEDIUMP_FLT_MAX)
const float PI = 3.14159265359;

struct BRDFParam {
    float metallic  ;
    float roughness ;
    float ao        ;
};

uniform vec4 colorParam;
uniform float selfColorIntensity;

#ifdef COLORMASKMAP
uniform sampler2D colorMaskMap;
#endif
 
vec3 colorComposite(vec3 compositeColor, vec3 inputColor) { 
    return inputColor;
}
 
#define NR_DIR_LIGHTS 2
uniform sampler2D brdfLut;
uniform samplerCube prefilterMap;
uniform samplerCube irradianceMap;
uniform float IBLIntensity;
uniform float IBLAngle;

vec3 FresnelSchlickRoughness(float cosTheta, vec3 F0, float roughness) {
    return F0 + (max(vec3(1.0 - roughness), F0) - F0) * pow(max(1.0 - cosTheta, 0.0), 5.0);
}

vec3 getIBLColor(vec3 N, vec3 V, vec3 albedo, BRDFParam p) {
    float angle = IBLAngle * 3.14;
    mat4 rotateMat = mat4(cos(angle), 0.0, sin(angle), 0.0, 0.0, 1.0, 0.0, 0.0, -sin(angle), 0.0, cos(angle), 0.0, 0.0, 0.0, 0.0, 1.0);
    vec3 view = V;        
    
    float roughness =  sqrt(p.roughness);
    roughness = clamp(roughness, 0.0, 1.0);

    vec3 F0 = vec3(fresnelIntense);
    F0 = mix(F0, albedo.rgb, p.metallic);
    vec3 F = FresnelSchlickRoughness(max(dot(N, view), 0.0), F0, roughness);

    vec3 kS = F;
    vec3 kD = 1.0 - kS;
    kD *= 1.0 - p.metallic;     
    vec3 R = reflect(-view, N);   
    R = (rotateMat * vec4(R, 1.0)).rgb;

    vec3 normal = (rotateMat * vec4(N, 1.0)).rgb;

    vec3 irradiance = textureCube(irradianceMap, normal).rgb;
    vec3 diffuse    = irradiance * albedo;

    const float MAX_REFLECTION_LOD = 7.0;
    vec3 prefilteredColor = textureCube(prefilterMap, R,  roughness * MAX_REFLECTION_LOD).rgb;
    float preBRDFX = clamp(dot(N, view), 0.01, 0.99);
    float preBRDFY = clamp(1.0 - roughness, 0.01, 0.99);
    vec2 envBRDF  = texture2D(brdfLut, vec2(preBRDFX, preBRDFY)).rg;
    vec3 specular = prefilteredColor * (F * envBRDF.x + envBRDF.y);

    vec3 ambient = (kD * diffuse + specular)* IBLIntensity * p.ao;
    return ambient;
}

vec3 getIBLColorClearCoat(vec3 N, vec3 V, vec3 albedo, BRDFParam p, float clearCoatPower, float clearCoatRoughness) {
    float angle = IBLAngle * 3.14;
    mat4 rotateMat = mat4(cos(angle), 0.0, sin(angle), 0.0, 0.0, 1.0, 0.0, 0.0, -sin(angle), 0.0, cos(angle), 0.0, 0.0, 0.0, 0.0, 1.0);
    vec3 view = V;        
    
    float roughness =  sqrt(p.roughness);
    roughness = clamp(roughness, 0.0, 1.0);

    vec3 F0 = vec3(fresnelIntense);
    F0 = mix(F0, albedo.rgb, p.metallic);
    vec3 F = FresnelSchlickRoughness(max(dot(N, view), 0.0), F0, roughness);

    vec3 kS = F;
    vec3 kD = 1.0 - kS;
    kD *= 1.0 - p.metallic;     
    vec3 R = reflect(-view, N);   
    R = (rotateMat * vec4(R, 1.0)).rgb;

    vec3 normal = (rotateMat * vec4(N, 1.0)).rgb;

    vec3 irradiance = textureCube(irradianceMap, normal).rgb;
    vec3 diffuse    = irradiance * albedo;

    const float MAX_REFLECTION_LOD = 7.0;
    vec3 prefilteredColor = textureCube(prefilterMap, R,  roughness * MAX_REFLECTION_LOD).rgb;
    float preBRDFX = clamp(dot(N, view), 0.01, 0.99);
    float preBRDFY = clamp(1.0 - roughness, 0.01, 0.99);
    vec2 envBRDF  = texture2D(brdfLut, vec2(preBRDFX, preBRDFY)).rg;
    vec3 specular = prefilteredColor * (F * envBRDF.x + envBRDF.y);

    float roughnessCoat = sqrt(clearCoatRoughness);
    roughnessCoat = clamp(roughnessCoat, 0.0, 1.0);

    vec3 prefilteredCoatColor = textureCube(prefilterMap, R,  roughnessCoat * MAX_REFLECTION_LOD).rgb;

    vec3 Fc = FresnelSchlickRoughness(max(dot(N, view), 0.0), vec3(0.04), roughnessCoat);
    float preBRDFXCoat = clamp(dot(N, view), 0.01, 0.99);
    float preBRDFYCoat = clamp(1.0 - roughnessCoat, 0.01, 0.99);
    vec2 envBRDFCoat  = texture2D(brdfLut, vec2(preBRDFXCoat, preBRDFYCoat)).rg;
    vec3 specularCoat = prefilteredCoatColor * (Fc * envBRDFCoat.x + envBRDFCoat.y) * clearCoatPower; 
    
    float attenuation = 1.0 - Fc.x*clearCoatPower;

    vec3 ambient = (kD * diffuse + specular * attenuation * attenuation + specularCoat)* IBLIntensity * p.ao;
    return ambient;
}

vec3 getIBLColorIridescence(vec3 N, vec3 V, vec3 albedo, BRDFParam p, float iridescenceFresnel, vec3 iridescenceF0, float iridescenceFactor) {
    float angle = IBLAngle * 3.14;
    mat4 rotateMat = mat4(cos(angle), 0.0, sin(angle), 0.0, 0.0, 1.0, 0.0, 0.0, -sin(angle), 0.0, cos(angle), 0.0, 0.0, 0.0, 0.0, 1.0);
    vec3 view = V;
    
    float roughness =  sqrt(p.roughness);
    roughness = clamp(roughness, 0.0, 1.0);

    vec3 F0 = vec3(fresnelIntense);
    // F0 = mix(F0, albedo.rgb, p.metallic); // need to be deleted
    vec3 F = FresnelSchlickRoughness(max(dot(N, view), 0.0), F0, roughness);
    
    // specular with iridescence
    vec3 kS = mix(F, vec3(iridescenceFresnel), iridescenceFactor);

    vec3 R = reflect(-view, N);   
    R = (rotateMat * vec4(R, 1.0)).rgb;

    const float MAX_REFLECTION_LOD = 7.0;
    vec3 prefilteredColor = textureCube(prefilterMap, R,  roughness * MAX_REFLECTION_LOD).rgb;
    float preBRDFX = clamp(dot(N, view), 0.01, 0.99);
    float preBRDFY = clamp(1.0 - roughness, 0.01, 0.99);
    vec2 envBRDF  = texture2D(brdfLut, vec2(preBRDFX, preBRDFY)).rg;
    vec3 specular = prefilteredColor * (kS * envBRDF.x + envBRDF.y);

    // diffuse with iridescence
    vec3 iridescenceF0Max = vec3(max(max(iridescenceF0.r, iridescenceF0.g), iridescenceF0.b));
    vec3 mixedF0 = mix(F0, iridescenceF0Max, iridescenceFactor);

    F = FresnelSchlickRoughness(max(dot(N, view), 0.0), mixedF0, roughness);

    vec3 kD = 1.0 - F;
    kD *= 1.0 - p.metallic;

    vec3 normal = (rotateMat * vec4(N, 1.0)).rgb;
    vec3 irradiance = textureCube(irradianceMap, normal).rgb;
    vec3 diffuse    = irradiance * albedo;

    vec3 ambient = (kD * diffuse + specular * IBLIntensity) * p.ao;
    return ambient;
}
//data struct
#define MAX_LIGHT_NUM 8
struct DirLight {
    vec3 position;
    vec3 direction;
    vec3 intensity;
    vec3 intenseColor;
};

struct PointLight {
    vec3 position;
    vec3 intensity;
};

struct SpotLight {
    vec3 position;
    vec3 direction;
    vec3 intensity;
    float cutOff;
    float outerCutOff;
};

struct PolygonalLight {
    vec3 positions[4];
    float intensity;
    vec3 intenseColor;
    bool doubleSided;
};

//global param
int lightNum = 0;
Light lights[MAX_LIGHT_NUM];

#ifdef NR_DIR_LIGHTS
	uniform DirLight dirLights[NR_DIR_LIGHTS];
#endif

#ifdef NR_POINT_LIGHTS
	uniform PointLight pointLights[NR_POINT_LIGHTS];
#endif

#ifdef NR_SPOT_LIGHTS
	uniform SpotLight spotLights[NR_SPOT_LIGHTS];
#endif

#ifdef NR_POLYGONAL_LIGHTS
    uniform PolygonalLight polygonalLights[NR_POLYGONAL_LIGHTS];
#endif

//struct light was definated in main BRDF
Light parseDirLight(vec3 worldPosition, DirLight dirLight) {
    Light ans;
    ans.direction = dirLight.direction;
    ans.intensity = dirLight.intensity * dirLight.intenseColor;
    float attenuation;
    //vWorldPosition comes from main brdf
    float distance = length(dirLight.position - worldPosition);
    attenuation = 1.0 / 5000.0;// / (1.0/*constant*/ + 0.09/*linear*/ * distance + 0.032/*quadratic*/ * (distance * distance));
    ans.attenuation = attenuation;
    return ans;
}

Light parsePointLight(vec3 worldPosition, PointLight pointLight) {
    Light ans;
    ans.direction = normalize(pointLight.position - worldPosition);
    ans.intensity = pointLight.intensity;
    float attenuation;
    
    float distance = length(pointLight.position - worldPosition);
    attenuation = 1.0 / 5000.0;// / (1.0/*constant*/ + 0.09/*linear*/ * distance + 0.032/*quadratic*/ * (distance * distance));
    ans.attenuation = attenuation;
    return ans;
}

Light parseSpotLight(vec3 worldPosition, SpotLight spotLight) {
    Light ans;
    ans.direction = spotLight.direction;

    vec3 lightDir = normalize(spotLight.position - worldPosition);
    float theta     = dot(lightDir, normalize(-spotLight.direction));
    float epsilon   = spotLight.cutOff - spotLight.outerCutOff;
    float intensity = clamp((theta - spotLight.outerCutOff) / epsilon, 0.0, 1.0);
    ans.intensity = intensity * spotLight.intensity;

    float attenuation;

    float distance = length(spotLight.position - vWorldPosition);
    attenuation = 1.0 / 5000.0 ;// / (1.0/*constant*/ + 0.09/*linear*/ * distance + 0.032/*quadratic*/ * (distance * distance));
    ans.attenuation = attenuation;
    return ans;
}


void getLights(vec3 worldPosition) {
    int tmpNum = 0;
#ifdef NR_DIR_LIGHTS
    for (int i = 0; i < NR_DIR_LIGHTS; i++) {
        lights[tmpNum] = parseDirLight(worldPosition, dirLights[i]);
        tmpNum++;
    }
    lightNum += NR_DIR_LIGHTS;
#endif
#ifdef NR_POINT_LIGHTS
    for (int i = 0; i < NR_POINT_LIGHTS; i++) {
        lights[tmpNum] = parsePointLight(worldPosition, pointLights[i]);
        tmpNum++;
    }
    lightNum += NR_POINT_LIGHTS;
#endif
#ifdef NR_SPOT_LIGHTS
    for (int i = 0; i < NR_SPOT_LIGHTS; i++) {
        lights[tmpNum] = parseSpotLight(worldPosition, spotLights[i]);
        tmpNum++;
    }
    lightNum += NR_SPOT_LIGHTS;
#endif
}
// uniform vec3 u_Coefficients[9];

float frac(float v)
{
    return v - floor(v);
}

vec2 frac(vec2 v)
{
    return v - floor(v);
}

vec2 delaunay_noise_randomVector(vec2 UV, float offset)
{
    mat2 m = mat2(15.27, 47.63, 99.41, 89.98);
    UV = frac(sin(UV * m) * 46839.32);
    return vec2(sin(UV.y * offset), cos(UV.x * offset)) * 0.25;
}

// Signed distance to a line crossing (p0, p1) segment.
float distLine(vec2 p0, vec2 p1)
{
    vec2 e0 = p1 - p0;
    return dot(p0, normalize(vec2(e0.y, -e0.x)));
}

// Use "Parabolic lifting" method to calculate if two triangles are about to flip.
// This is actually more reliable than circumscribed circle method.
// The technique is based on duality between Delaunay Triangulation
// and Convex Hull, where DT is just a boundary of convex hull
// of projected seeds onto paraboloid.
// We project (h1 h2 h3) triangle ontot paraboloid
// and return the distance of the origin
// to a plane crossing projected triangle.
float flipDistance(vec2 h1, vec2 h2, vec2 h3)
{
    // Project triangle on paraboloid.
    vec3 g1 = vec3(h1, dot(h1, h1));
    vec3 g2 = vec3(h2, dot(h2, h2));
    vec3 g3 = vec3(h3, dot(h3, h3));

    return dot(g1, cross(g3 - g1, g2 - g1));
}

// Find the distance to the closest Delaunay edge in quad (h0, h1, h2, h3).
void delaunayQuad(vec2 h0, vec2 h1, vec2 h2, vec2 h3, vec2 pos0, vec2 pos1, vec2 pos2, vec2 pos3, inout float distance, inout vec2 center)
{
    // Get the distance to quad edges.
    // note: in general it can be concave, but we don't care.
    float minDistance = min(min(distLine(h0, h1), distLine(h1, h2)), min(distLine(h2, h3), distLine(h3, h0)));

    if (minDistance < 0.0) return; // Outside the quad

    // Calculate flip distance relative to h2.
    float dc = flipDistance(h0 - h2, h1 - h2, h3 - h2);

    // Flipping rotates diagonal from (h0 h2) to (h3 h1).
    vec2 diagonalPosition0 = (dc > 0.0) ? h3 : h0;
    vec2 diagonalPosition1 = (dc > 0.0) ? h1 : h2;

    // Calculate distance to diagonal.
    float distToDiagonal = distLine(diagonalPosition0, diagonalPosition1);

    // Calculate distance to diagonal (positive from both sides).
    minDistance = min(minDistance, abs(distToDiagonal));

    distance = min(minDistance, distance);

    if (dc > 0.0)
    {
        if (distToDiagonal > 0.0)
        center = (pos1 + pos2 + pos3) / 3.0;
        else
        center = (pos0 + pos1 + pos3) / 3.0;
    }
    else
    {
        if (distToDiagonal > 0.0)
        center = (pos0 + pos2 + pos3) / 3.0;
        else
        center = (pos0 + pos1 + pos2) / 3.0;
    }
}

// Final function visits 4 quads around the center cell.
void delaunayTriangulation(vec2 UV, float AngleOffset, out float minDistance, out vec2 center)
{
    vec2 grid = floor(UV);

    // Cache sites coordinate.
    vec2 site[9];
    vec2 h[9];

    // Iterate sites.
    for (int y = -1; y <= 1; y++)
    {
        for (int x = -1; x <= 1; x++)
        {
            vec2 lattice = vec2(x, y);
            vec2 offset = vec2(0.5, 0.5) + delaunay_noise_randomVector(grid + lattice, AngleOffset);
            site[(x + 1) + (y + 1) * 3] = grid + lattice + offset;
            h[(x + 1) + (y + 1) * 3] = site[(x + 1) + (y + 1) * 3] - UV;
        }
    }

    minDistance = 8.0;
    center = vec2(0.0, 0.0);

    delaunayQuad(h[3], h[0], h[1], h[4], site[3], site[0], site[1], site[4], minDistance, center);
    delaunayQuad(h[4], h[1], h[2], h[5], site[4], site[1], site[2], site[5], minDistance, center);
    delaunayQuad(h[7], h[4], h[5], h[8], site[7], site[4], site[5], site[8], minDistance, center);
    delaunayQuad(h[6], h[3], h[4], h[7], site[6], site[3], site[4], site[7], minDistance, center);
}

// Delaunay Triangluation
// https://www.shadertoy.com/view/4sKyRD
void delaunayTriangulationNoise(vec2 UV, float AngleOffset, float CellDensity, out float Out, out vec2 Cells)
{
    delaunayTriangulation(UV * CellDensity, AngleOffset, Out, Cells);
}

float getRandomValue(vec2 uv)
{
    return frac(sin(dot(uv.xy, vec2(12.9898, 78.233))) * 43758.5453);
}

// vec3 sampleSH(vec3 N)
// {
//     float basis[9];
//     float x = N.x;
//     float y = N.y;
//     float z = N.z;
//     float x2 = x * x;
//     float y2 = y * y;
//     float z2 = z * z;

//     basis[0] = 1.0 / 2.0 * sqrt(1.0 / PI) * PI;

//     basis[1] = -2.0 / 3.0 * sqrt(3.0 / (4.0 * PI)) * y * PI;
//     basis[2] = 2.0 / 3.0 * sqrt(3.0 / (4.0 * PI)) * z * PI;
//     basis[3] = -2.0 / 3.0 * sqrt(3.0 / (4.0 * PI)) * x * PI;

//     basis[4] = 1.0 / 4.0 * 1.0 / 2.0 * sqrt(15.0 / PI) * y * x * PI;
//     basis[5] = -1.0 / 4.0 * 1.0 / 2.0 * sqrt(15.0 / PI) * y * z * PI;
//     basis[6] = 1.0 / 4.0 * 1.0 / 4.0 * sqrt(5.0 / PI) * (3.0 * z2 - 1.0) * PI;
//     basis[7] = -1.0 / 4.0 * 1.0 / 2.0 * sqrt(15.0 / PI) * z * x * PI;
//     basis[8] = 1.0 / 4.0 * 1.0 / 4.0 * sqrt(15.0 / PI) * (x2 - y2) * PI;

//     vec3 irradiance = vec3(0.0, 0.0, 0.0);
//     for (int i = 0; i < 9; i++)
//         irradiance += u_Coefficients[i] * basis[i];

//     return irradiance;
// }

// vec3 highlights(vec3 P, float roughness, vec3 N, vec3 V)
// {
//     Light light = GetMainLight();
//     float roughness2 = roughness * roughness;
//     vec3 halfDir = normalize(light.direction + V);
//     float NoH = clamp(dot(normalize(N), halfDir), 0.0, 1.0);
//     float LoH = clamp(dot(light.direction, halfDir), 0.0, 1.0);
//     // GGX Distribution multiplied by combined approximation of Visibility and Fresnel
//     float d = NoH * NoH * (roughness2 - 1.h) + 1.0001h;
//     float LoH2 = LoH * LoH;
//     float specularTerm = roughness2 / ((d * d) * max(0.1, LoH2) * (roughness + 0.5) * 4);
//     specularTerm = min(specularTerm, 10);
//     return specularTerm * light.intensity * light.attenuation;
// }
 

varying vec4 FragPosLightSpace;
varying vec3 FragNormalLightSpace;
uniform float minBias;
uniform float maxBias;
uniform float shadowRadius;
uniform sampler2D depthMap;

float ShadowCalculation(vec4 fragPosLightSpace, vec3 lightDir)
{
    vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
    projCoords = projCoords * 0.5 + 0.5;
    float closestDepth = texture2D(depthMap, projCoords.xy).r;
    float currentDepth = projCoords.z;
    if (projCoords.x > 1.0 || projCoords.x < 0.0 || projCoords.y > 1.0 || projCoords.y < 0.0 || projCoords.z > 1.0 || projCoords.z < 0.0) {
        return 0.0;
    }
    vec3 L = (lightDir);
    vec3 N = normalize(FragNormalLightSpace);
    float bias = max(maxBias * (1.0 - dot(N, L)), minBias);
 
    // float shadow = (currentDepth - bias) > closestDepth  ? 1.0 : 0.0;
    
    // PCF
    float shadow = 0.0;
    vec2 texelSize = shadowRadius / vec2(2048.0);
    for(int x = -1; x <= 1; ++x)
    {
        for(int y = -1; y <= 1; ++y)
        {
            float pcfDepth = texture2D(depthMap, projCoords.xy + vec2(x, y) * texelSize).r; 
            shadow += currentDepth - bias > pcfDepth  ? 1.0 : 0.0;        
        }    
    }
    shadow /= 9.0;

    return shadow;
}

vec3 getNormalFromMap(){
    vec3 tangentNormal = texture2D(normalMap, vTexcoord).rgb * 2.0 - vec3(1.0);

    vec3 N = normalize(vNormal);
    vec3 T = normalize(vTangent);
    vec3 B = normalize(cross(N,T));
    mat3 TBN = mat3(T,B,N);
    return normalize(TBN * normalize(tangentNormal));
}

//cosTheta : n dot v
vec3 fresnelSchlick(float cosTheta, vec3 F0){
    return F0 + (1.0 - F0) * pow(max(1.0 - cosTheta, 0.0), 5.0);
}

float VisibilityInMarmoset(float NoV, float NoL, float a) { 
	float k = a * 0.5;
	float G_SmithL = NoL * (1.0 - k) + k;
	float G_SmithV = NoV * (1.0 - k) + k;
	float G = 0.25 / ( G_SmithL * G_SmithV );
    return saturateMediump(G);
}

float D_GGX(const vec3 n, const vec3 h, float roughness) {
    vec3 NxH = cross(n, h);
    float NoH = dot(n,h);
    float a = NoH * roughness;
    float k = roughness / (dot(NxH, NxH) + a * a);
    float d = k * k * (1.0 / PI);
    return saturateMediump(d);
}

const vec2 invAtan = vec2(0.1591, 0.3183);
vec2 SampleSphericalMap(vec3 v)
{
    vec2 uv = vec2(atan(v.z, v.x), asin(v.y));
    uv *= invAtan;
    uv += 0.5;
    return uv;
}

vec3 BRDF(Light light, BRDFParam p, vec3 N, vec3 V, vec3 albedo) {
    vec3 color = vec3(0.0);
    vec3 F0 = vec3(fresnelIntense);
    F0 = mix(F0, albedo.rgb, p.metallic);
    
    vec3 L = -light.direction;
    vec3 H = normalize(V + L);
    float NoV = clamp(dot(N, V), 0.0, 1.0);
    float NoL = clamp(dot(N, L), 0.0, 1.0);
    float LoH = clamp(dot(H, L), 0.0, 1.0);    
    float VoH = clamp(dot(V, H), 0.0, 1.0);

//base brdf
//pre calculate of Standard brdf
    vec3 radiance = light.attenuation * light.intensity;
    vec3 FSchlick   = fresnelSchlick(LoH, F0); 
    vec3 F = FSchlick;

    //vec3 dfg = texture2D(brdfLut, vec2(clamp(NoV, 0.01, 0.99), clamp(1.0 - p.roughness, 0.01, 0.99))).rgb;
    vec3 energyCompensation = vec3(1.0);// + F0 * (1.0 / dfg.y - 1.0);

    // specular
#ifdef IRIDESCENCE    
    solveIridescenceFresnel(dot(N, V), F0);
    F = mix(FSchlick, iridescenceFresnel, iridescenceFactor);
#endif

#ifdef ANISOTROPIC
    float NDFG = getAnitropicMicroNormal(N, L, H, V, p.roughness);
    vec3 specular = NDFG * F * specularIntensity * energyCompensation;
#else
    float NDF = D_GGX(N, H, p.roughness);
    float G   = VisibilityInMarmoset(NoV, NoL, p.roughness);
    vec3 nominator = NDF * G * F;
    vec3 specular = nominator * specularIntensity * energyCompensation;
#endif
    
//diffuse
#ifdef IRIDESCENCE
    vec3 iridescenceF0Max = vec3(max(max(iridescenceF0.r, iridescenceF0.g), iridescenceF0.b));
    F = mix(FSchlick, iridescenceF0Max, iridescenceFactor);
#endif

    vec3 kS = F;
    vec3 kD = vec3(1.0) - kS;
    kD *= 1.0 - p.metallic;
    float NdotL = max(dot(N, L), 0.0);
    vec3 diffuseColor = kD * albedo.rgb / PI * (NdotL + wrapped) / ((1.0 + wrapped) * (1.0 + wrapped));
#ifdef SHEEN     
    float NoH = clamp(dot(N, H), 0.0, 1.0);
    diffuseColor *= sheenScaling;
    specular *= sheenScaling;
    specular += getSheenColor(NoL, NoV, NoH);
#endif

#ifdef CLEARCOAT
    float Frc = getClearCoatPower(N, H, V, L);
    float Fc = fresnelSchlick(LoH, vec3(0.04)).x * clearCoatPower;
    color += (specular * (1.0 - Fc) * (1.0 - Fc) + Frc) * NdotL + diffuseColor;
#else
    color += specular * NdotL + diffuseColor;
#endif

    color = color * radiance;

#ifdef SUBSURFACE
    color += getSubsurfaceColor(kD, radiance, V, -light.direction, N, albedo);
#endif

    return color;
}

uniform float metallicParam;
uniform float roughnessParam;
void main()
{
#ifdef MESH_MASK
    if (vMask > 0.99999) {
        discard;
    }
#endif
    
    vec4 mapColor = texture2D(diffuseMap, vTexcoord).rgba;
    vec3 albedo = mapColor.rgb; 
    if (colorParam.w > 0.0) {
        float colorMask = 1.0;
#ifdef COLORMASKMAP
        colorMask = texture2D(colorMaskMap, vTexcoord).r;
#endif 
        albedo = colorMask * colorComposite(colorParam.rgb, albedo) + (1.0 - colorMask) * albedo;        
    }
    albedo = pow(albedo, vec3(2.2)); 
    getLights(vWorldPosition);
    BRDFParam p;
    BRDFParam IBLBrdfParam;
    vec3 mraColor   = texture2D(mraMap,vTexcoord).rgb;
#ifdef METALLICPARAM
    p.metallic  = clamp(metallicParam, 0.01, 1.0);
#else
    if (metallicParam < 1.0) {
        p.metallic  = (mraColor.r * metallicParam);
    } else {
        p.metallic  = (mraColor.r) + (metallicParam - 1.0) / 2.0 * (1.0 - mraColor.r);
    }
    p.metallic = clamp(p.metallic , 0.01, 1.0);
#endif

    IBLBrdfParam.metallic = p.metallic;
#ifdef ROUGHNESSPARAM
    p.roughness =  roughnessParam;
#else
    if (roughnessParam < 1.0) {
        p.roughness = (mraColor.g) * roughnessParam;
    } else {
        p.roughness = (mraColor.g) + (roughnessParam - 1.0) / 2.0 * (1.0 - mraColor.g);
    }
#endif
    IBLBrdfParam.roughness = p.roughness;
    p.roughness = clamp(p.roughness * p.roughness, 0.01, 1.0);

    p.ao        = mraColor.b;
    IBLBrdfParam.ao = p.ao;

#ifdef NORMALMAP 
    vec3 N = getNormalFromMap();
#else
    vec3 N = normalize(vNormal);
#endif
#ifdef DOUBLESIDED
    if (!gl_FrontFacing) {
        N = -N;
    }
#endif
    vec3 V = normalize(cameraPosition - vWorldPosition);
    vec3 R = reflect(-V, N);
    
    vec3 Lo = vec3(0.0);

#ifdef LASER
    albedo = getLaser(N, V, albedo);
#endif

//base brdf
//pre calculate of Standard brdf
float shadow = 0.0;
    for (int i = 0; i < lightNum; i++) {
#ifdef SHADOW
        if (i == 0) {
            shadow = ShadowCalculation(FragPosLightSpace, lights[0].direction);
            Lo += (1.0 - shadow) * BRDF(lights[i], p, N, V, albedo);
        } else {
            Lo += BRDF(lights[i], p, N, V, albedo);
        }
#else
        Lo += BRDF(lights[i], p, N, V, albedo);
#endif
#ifdef RIMLIGHT
        Lo += rimLight(V, -lights[i].direction, N, albedo); 
#endif
    } 

#ifdef IBL
#ifdef IRIDESCENCE
    Lo += getIBLColorIridescence(N, V, albedo, IBLBrdfParam, iridescenceFresnel, iridescenceF0, iridescenceFactor);
#elif defined CLEARCOAT
    Lo += getIBLColorClearCoat(N, V, albedo, IBLBrdfParam, clearCoatPower, clearCoatRoughness);
#else
    Lo += getIBLColor(N, V, albedo, IBLBrdfParam);
#endif
#endif

#ifdef LTC
    #ifdef NR_POLYGONAL_LIGHTS
        for (int i = 0; i < NR_POLYGONAL_LIGHTS; i++) {
            Lo += getLTCColor(normalize(vNormal), V, vWorldPosition, polygonalLights[i], albedo, IBLBrdfParam.metallic, IBLBrdfParam.roughness);
        }
    #endif    
#endif

//self color
    Lo += selfColorIntensity * albedo;
    vec3 color = Lo;//ambient + Lo;
    color = pow(color , vec3(1.0 / 2.2));
    
    gl_FragColor = vec4(color, mapColor.a * opacity);
}
